package cs108;

import static java.lang.Math.cos;
import static java.lang.Math.floor;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.lang.Math.sin;
import static java.lang.Math.sqrt;

/**
 * Une image continue et infinie, représentée par une fonction associant une
 * valeur d'un type donné (p.ex. une couleur) à chaque point du plan.
 */

public interface Image<T> {
    public T apply(float x, float y);

    public static final class RedDisk implements Image<ColorRGB> {
        @Override
        public ColorRGB apply(float x, float y) {
            double r = sqrt(x * x + y * y);
            return r <= 1d ? ColorRGB.RED : ColorRGB.WHITE;
        }
    }

    public static final Image<ColorRGB> RED_DISK = new RedDisk();

    public static final class Chessboard implements Image<ColorRGB> {
        private final ColorRGB c1, c2;
        private final float w;

        public Chessboard(ColorRGB c1, ColorRGB c2, float w) {
            if (! (w > 0))
                throw new IllegalArgumentException("non-positive width: " + w);
            this.c1 = c1;
            this.c2 = c2;
            this.w = w;
        }

        @Override
        public ColorRGB apply(float x, float y) {
            int sqX = (int)floor(x / w), sqY = (int)floor(y / w);
            return (sqX + sqY) % 2 == 0 ? c1 : c2;
        }
    }

    public static final class HorizontalGradientMask implements Image<Float> {
        @Override
        public Float apply(float x, float y) {
            return max(0, min((x + 1f) / 2f, 1f));
        }
    }

    public static final Image<Float> HORIZONTAL_GRADIENT_MASK = new HorizontalGradientMask();

    public static final class Composed implements Image<ColorRGB> {
        private final Image<ColorRGB> bg, fg;
        private final Image<Float> mask;

        public Composed(Image<ColorRGB> bg, Image<ColorRGB> fg, Image<Float> mask) {
            this.bg = bg;
            this.fg = fg;
            this.mask = mask;
        }

        @Override
        public ColorRGB apply(float x, float y) {
            return bg.apply(x, y).mixWith(fg.apply(x, y), mask.apply(x, y));
        }
    }

    public static final class Mysterious<T> implements Image<T> {
        private final Image<T> image;

        public Mysterious(Image<T> image) {
            this.image = image;
        }

        @Override
        public T apply(float x, float y) {
            return image.apply(x * 2, y);
        }
    }

    public static final class Rotated<T> implements Image<T> {
        private final Image<T> image;
        private final float cosA, sinA;

        public Rotated(Image<T> image, float angle) {
            this.image = image;
            this.cosA = (float) cos(-angle);
            this.sinA = (float) sin(-angle);
        }

        @Override
        public T apply(float x, float y) {
            float x1 = x * cosA - y * sinA;
            float y1 = x * sinA + y * cosA;
            return image.apply(x1, y1);
        }
    }
}
